/***************************************************************************
 *   Copyright (C) 2007 by Daniel Guernsey                                 *
 *   danpg102@hotmail.com                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


//#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>

#include <GL/glx.h>
//#include <SDL/SDL.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "vroot.h"

int GRID_W = 64;
int GRID_H = 48;

int drawroot = 0;

float mousepos[2];

int windowW = 1280;
int windowH = 1024;

float step = 0.0003;
int x_freq = 10;
int y_freq = 21;
float w_factor = 0.9;
float h_factor = 0.9;

float force = 60;
float threshhold = 10;
float damp_factor = 0.995;

float color_back[] = {0, 0, 0, 0};

// Fire
float h_color_const[] = {1, 0, 0, 0};
float h_color_factor[] = {0, 0.2, 0, 1};
float v_color_const[] = {1, 0, 0, 0};
float v_color_factor[] = {0, 0.2, 0, 1};

// Electricity
/*
float h_color_const[] = {0, 0, 1, 0};
float h_color_factor[] = {0.2, 1, 0, 1};
float v_color_const[] = {0, 0, 1, 0};
float v_color_factor[] = {0.2, 1, 0, 1};
*/

//float sprites[GRID_W][GRID_H][2];
//float vel[GRID_W][GRID_H][2];

float*** sprites;
float*** vel;


// GLX
Window window = 0;
Display *xdisplay;
int screen;


//SDL_Surface* screen;

void printusage(char* progname) {
  fprintf(stderr, "Tempest screensaver for X\n");
  fprintf(stderr, "Usage: %s OPTIONS\n", progname);
  fprintf(stderr, "  Options:\n");
  fprintf(stderr, "    -root        Draw in the root window\n");
  fprintf(stderr, "    -window-id   Draw in the named window\n");
  fprintf(stderr, "    -gridw       The number of columns of sprites\n");
  fprintf(stderr, "    -gridh       The number of rows of sprites\n");
  fprintf(stderr, "    -step        The increment of movement for the force sprite\n");
  fprintf(stderr, "    -xfreq       The scale factor for horizontal movement of the force sprite\n");
  fprintf(stderr, "    -yfreq       The scale factor for vertical movement of the force sprite\n");
  fprintf(stderr, "    -wfactor     Defines the maximum width of movement of the force sprite\n");
  fprintf(stderr, "    -hfactor     Defines the maximum height of movement of the force sprite\n");
  fprintf(stderr, "    -force       How much for the force sprite applies on the others\n");
  fprintf(stderr, "    -threshhold  Minimum distance for a sprite to have force applied to it\n");
  fprintf(stderr, "    -damp        Percentage of velocity each sprite will lose each frame\n");
  fprintf(stderr, "    -backcolor   Background color (4-tuple floating point)\n");
  fprintf(stderr, "         params: RED GREEN BLUE ALPHA\n");
  fprintf(stderr, "    -color       Color of the grid lines connecting the sprites\n");
  fprintf(stderr, "         params: HORIZONALB HORIZONALM VERTICALB VERTICALM\n");
  fprintf(stderr, "           help: each parameter is a 4-tuple (like the -backcolor option)\n");
  fprintf(stderr, "                 color is calculated linearly C = Mf + B where f is the total\n");
  fprintf(stderr, "                   force being applied to that sprite\n");
}

void display() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  int i;
  int j;

  glPushMatrix();
    glTranslatef(-1.0, 1.0, 0);
    glScalef(2.0 / windowW, -2.0 / windowH, 1.0);
/*  glBegin(GL_QUADS);
      glColor3f(1, 0, 0);
      for (i = 0; i < GRID_W; i++) {
        for (j = 0; j < GRID_H; j++) {
          glVertex2f(sprites[i][j][0] - 10.0, sprites[i][j][1] - 10.0);
          glVertex2f(sprites[i][j][0] + 10.0, sprites[i][j][1] - 10.0);
          glVertex2f(sprites[i][j][0] + 10.0, sprites[i][j][1] + 10.0);
          glVertex2f(sprites[i][j][0] - 10.0, sprites[i][j][1] + 10.0);
        }
      }
    glEnd();*/

    float brightness;
    for (i = 0; i < GRID_W; i++) {
      glBegin(GL_LINE_STRIP);
        for (j = 0; j < GRID_H; j++) {
          brightness = sqrt(vel[i][j][0]*vel[i][j][0] + vel[i][j][1]*vel[i][j][1]);
            glColor4f(v_color_const[0] + v_color_factor[0] * brightness,
                      v_color_const[1] + v_color_factor[1] * brightness,
                      v_color_const[2] + v_color_factor[2] * brightness,
                      v_color_const[3] + v_color_factor[3] * brightness);
          glVertex2f(sprites[i][j][0], sprites[i][j][1]);
        }
      glEnd();
    }
    for (j = 0; j < GRID_H; j++) {
      glBegin(GL_LINE_STRIP);
        for (i = 0; i < GRID_W; i++) {
          brightness = sqrt(vel[i][j][0]*vel[i][j][0] + vel[i][j][1]*vel[i][j][1]);
            glColor4f(h_color_const[0] + h_color_factor[0] * brightness,
                      h_color_const[1] + h_color_factor[1] * brightness,
                      h_color_const[2] + h_color_factor[2] * brightness,
                      h_color_const[3] + h_color_factor[3] * brightness);
          glVertex2f(sprites[i][j][0], sprites[i][j][1]);
        }
      glEnd();
    }
  glPopMatrix();

  //glutSwapBuffers();
  //SDL_GL_SwapBuffers();
  glXSwapBuffers(xdisplay, window);
}

void idle() {
  static float accel[2];

  float dist[2];
  float distance;
  float edistance;
  
  static float stepper = 0;
  
  mousepos[0] = (1.0 * windowW) / 2 * (1.0 + w_factor * sin(stepper * x_freq));
  mousepos[1] = (1.0 * windowH) / 2 * (1.0 + h_factor * sin(stepper * y_freq));
  
  stepper+= step;
  if (stepper > 6.284) {
    stepper-= 6.284;
  }
  if (stepper < -6.284) {
    stepper+= 6.284;
  }
  
  int i;
  int j;
  int k;
  int l;
  for (i = 1; i < GRID_W - 1; i++) {
    for (j = 1; j < GRID_H - 1; j++) {
      accel[0] = 0;
      accel[1] = 0;
      dist[0] = mousepos[0] - sprites[i][j][0];
      dist[1] = mousepos[1] - sprites[i][j][1];
      distance = sqrt(dist[0]*dist[0] + dist[1]*dist[1]);
      
      edistance = sqrt(dist[0]*dist[0] + dist[1]*dist[1])/force;

      
      if (distance > threshhold) {
        accel[0]-= dist[0] / (edistance*edistance*edistance); // once for normal, twice more for inverse square
        accel[1]-= dist[1] / (edistance*edistance*edistance);
/*      if (accel[0] > 100) accel[0] = 100;
        if (accel[0] <-100) accel[0] =-100;
        if (accel[1] > 100) accel[1] = 100;
        if (accel[1] <-100) accel[1] =-100;*/
      }
      for (k = i - 1; k <= i + 1; k++) {
        for (l = j - 1; l <= j + 1; l++) {
          if (!(k == i && l == j)) {
            dist[0] = sprites[k][l][0] - sprites[i][j][0];
            dist[1] = sprites[k][l][1] - sprites[i][j][1];
//            distance = sqrt(dist[0]*dist[0] + dist[1]*dist[1]);
//            accel[0]+= dist[0] - windowW/10 * dist[0] / distance; // (distance - normal)
//            accel[1]+= dist[1] - windowH/10 * dist[1] / distance;
            accel[0]+= dist[0];
            accel[1]+= dist[1];
          }
        }
      }

      vel[i][j][0]+= accel[0]/1000;
      vel[i][j][1]+= accel[1]/1000;

      vel[i][j][0]*= damp_factor;
      vel[i][j][1]*= damp_factor;
    }
  }
  for (i = 0; i < GRID_W; i++) {
    for (j = 0; j < GRID_H; j++) {
      sprites[i][j][0]+= vel[i][j][0];
      sprites[i][j][1]+= vel[i][j][1];
    }
  }
  //glutPostRedisplay();
}

/*
void amouse(int x, int y) {
  mousepos[0] = x;
  mousepos[1] = y;
  mousedwn = 1;
}

void pmouse(int x, int y) {
  mousepos[0] = x;
  mousepos[1] = y;
  mousedwn = 0;
}

void mouseclick(int button, int state, int x, int y) {
  //glutPushWindow();
  if (state == GLUT_DOWN) {
    mousedwn = 1;
  } else {
    mousedwn = 0;
  }
}
*/

void free_stuff() {
  int i, j;
  fprintf(stderr, "Freeing crap\n");
  for (i = 0; i < GRID_W; i++) {
    for (j = 0; j < GRID_H; j++) {
      free(sprites[i][j]);
      free(vel[i][j]);
    }
    free(sprites[i]);
    free(vel[i]);
  }
  free(sprites);
  free(vel);
}

void reshape(int width, int height) {
  //glutPushWindow();
  windowW = width;
  windowH = height;
  glViewport(0, 0, windowW, windowH);

  int i;
  int j;
  for (i = 0; i < GRID_W; i++) {
    for (j = 0; j < GRID_H; j++) {
      sprites[i][j][0] = i * windowW / (GRID_W - 1);
      sprites[i][j][1] = j * windowH / (GRID_H - 1);
    }
  }
}

void parse(int argc, char** argv) {
  int i;
  for (i = 1; i < argc; i++) {
    if (!strncmp(argv[i], "-root", 5)) {
      drawroot = 1;
    } else if (!strncmp(argv[i], "-window-id", 10)) {
      i++;
      if (i < argc) {
        window = atoi(argv[i]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-gridw", 6)) {
      i++;
      if (i < argc) {
        GRID_W = atoi(argv[i]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-gridh", 6)) {
      i++;
      if (i < argc) {
        GRID_H = atoi(argv[i]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-step", 5)) {
      i++;
      if (i < argc) {
        //step = strtof(argv[i]);
        sscanf(argv[i], "%f", &step);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-xfreq", 6)) {
      i++;
      if (i < argc) {
        x_freq = atoi(argv[i]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-yfreq", 6)) {
      i++;
      if (i < argc) {
        y_freq = atoi(argv[i]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-wfactor", 8)) {
      i++;
      if (i < argc) {
        //w_factor = strtof(argv[i]);
        sscanf(argv[i], "%f", &w_factor);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-hfactor", 8)) {
      i++;
      if (i < argc) {
        //h_factor = strtof(argv[i]);
        sscanf(argv[i], "%f", &h_factor);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-force", 6)) {
      i++;
      if (i < argc) {
        //force = strtof(argv[i]);
        sscanf(argv[i], "%f", &force);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-threshhold", 11)) {
      i++;
      if (i < argc) {
        //threshhold = strtof(argv[i]);
        sscanf(argv[i], "%f", &threshhold);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-damp", 5)) {
      i++;
      if (i < argc) {
        //damp_factor = strtof(argv[i]);
        sscanf(argv[i], "%f", &damp_factor);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-backcolor", 10)) {
      i++;
      if (i < argc) {
        //color_back[0] = strtof(argv[i]);
        sscanf(argv[i], "%f", &color_back[0]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        //color_back[1] = strtof(argv[i]);
        sscanf(argv[i], "%f", &color_back[1]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        //color_back[2] = strtof(argv[i]);
        sscanf(argv[i], "%f", &color_back[2]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        //color_back[3] = strtof(argv[i]);
        sscanf(argv[i], "%f", &color_back[3]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else if (!strncmp(argv[i], "-color", 6)) {
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_const[0]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_const[1]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_const[2]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_const[3]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_factor[0]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_factor[1]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_factor[2]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &h_color_factor[3]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_const[0]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_const[1]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_const[2]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_const[3]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_factor[0]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_factor[1]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_factor[2]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
      i++;
      if (i < argc) {
        sscanf(argv[i], "%f", &v_color_factor[3]);
      } else {
        printusage(argv[0]);
        exit(1);
      }
    } else {
      fprintf(stderr, "Unknown option: %s\n", argv[i]);
      printusage(argv[0]);
      exit(1);
    }
  }
}

int main(int argc, char** argv) {
  parse(argc, argv);
  
  int i, j;
  sprites = malloc(sizeof(float**) * GRID_W);
  vel = malloc(sizeof(float**) * GRID_W);
  for (i = 0; i < GRID_W; i++) {
    sprites[i] = malloc(sizeof(float*) * GRID_H);
    vel[i] = malloc(sizeof(float*) * GRID_H);
    for (j = 0; j < GRID_H; j++) {
      sprites[i][j] = malloc(sizeof(float) * 2);
      vel[i][j] = malloc(sizeof(float) * 2);
    }
  }
    
  
  mousepos[0] = 0;
  mousepos[1] = 0;
  //GLUT
  //glutInit(&argc, argv);
  //glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);
  //glutInitWindowSize(windowW, windowH);
  
  if (!(xdisplay = XOpenDisplay(0))) {
    fprintf(stderr, "Could not open display\n");
    return -1;
  }
  screen = DefaultScreen(xdisplay);
  
  if (!window) {
    Window root = RootWindow(xdisplay, screen);
    
    if (drawroot) {
      window = root;
    } else {
      window = XCreateSimpleWindow(xdisplay, root, 0, 0, windowW, windowH, 0, 0, 0);
    }
  }
  
  XWindowAttributes attribs;
  XGetWindowAttributes(xdisplay, window, &attribs);
  Visual* visual = attribs.visual;
  
  XVisualInfo vi;
  vi.screen = screen;
  vi.visualid = XVisualIDFromVisual(visual);
  int out_count;
  XVisualInfo* vinfo = XGetVisualInfo(xdisplay, VisualScreenMask | VisualIDMask,
                                      &vi, &out_count);
  if (!vinfo) {
    return -1;
  }
  
  GLXContext context = glXCreateContext(xdisplay, vinfo, 0, GL_TRUE);
  XFree(vinfo);
  if (!context) {
    return -1;
  }
  
  if (!glXMakeCurrent(xdisplay, window, context)) {
    return -1;
  }
  
  XMapRaised(xdisplay, window);
  
  // SDL
  /*
  if (SDL_Init(SDL_INIT_VIDEO)) {
    fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
    return -1;
  }
  */
  
  //screen = SDL_SetVideoMode(windowW, windowH, 0, SDL_OPENGL | SDL_RESIZABLE | SDL_FULLSCREEN);
  
  // GLUT
  //unsigned int WINDOW_ID;
  //WINDOW_ID = glutCreateWindow("Tempest");
  //glutFullScreen();

  //glutDisplayFunc(display);
  //glutIdleFunc(idle);
  //glutMotionFunc(amouse);
  //glutPassiveMotionFunc(pmouse);
  //glutMouseFunc(mouseclick);
  //glutReshapeFunc(reshape);

  glClearColor(color_back[0], color_back[1], color_back[2], color_back[3]);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  //glutMainLoop();
  
  unsigned char quit = 0;
  
  //int mouseintpos[2];
  
  XGetWindowAttributes(xdisplay, window, &attribs);
  reshape(attribs.width, attribs.height);
//  reshape(windowW, windowH);
  
  atexit(free_stuff);
  
  while (!quit) {
    idle();
    display();
    
    /*
    SDL_Event event;
    while (SDL_PollEvent(&event)) {
      if (event.type == SDL_KEYDOWN) {
        quit = 1;
      } else if (event.type == SDL_QUIT) {
        quit = 1;
      } else if (event.type == SDL_VIDEORESIZE) {
        windowW = event.resize.w;
        windowH = event.resize.h;
        SDL_SetVideoMode(windowW, windowH, 0, SDL_OPENGL | SDL_RESIZABLE | SDL_FULLSCREEN);
        reshape(windowW, windowH);
      }
    }
    */
    
    XEvent event;
    while (XPending(xdisplay) && !quit) {
      fprintf(stderr, "%d\n", XPending(xdisplay));
      XNextEvent(xdisplay, &event);
      switch (event.type) {
        case (ConfigureNotify): {
          XWindowAttributes attribs;
          XGetWindowAttributes(xdisplay, window, &attribs);
          reshape(attribs.width, attribs.height);
        } break;
        case (DestroyNotify): {
          quit = 1;
        } break;
        case (MotionNotify): {
          mousepos[0] = event.xmotion.x;
          mousepos[1] = event.xmotion.y;
        } break;
      }
    }
    
    //mousedwn = SDL_GetMouseState(&mouseintpos[0], &mouseintpos[1]);
    //mousepos[0] = mouseintpos[0]; // convert int to float
    //mousepos[1] = mouseintpos[1];
  }

  //glutDestroyWindow(WINDOW_ID);
  
  //SDL_Quit();
  
  return 0;
}
